package com.cesgroup.bdc.wfself.engine.service.impl;

import com.cesgroup.bdc.wfself.def.CommonNode;
import com.cesgroup.bdc.wfself.def.CommonProcDef;
import com.cesgroup.bdc.wfself.def.INode;
import com.cesgroup.bdc.wfself.def.ITask;
import com.cesgroup.bdc.wfself.engine.constant.ProcConst;
import com.cesgroup.bdc.wfself.engine.service.WorkflowDao;
import com.cesgroup.bdc.wfself.entity.ProcInstHisInfo;
import com.cesgroup.bdc.wfself.entity.ProcInstRunVar;
import com.cesgroup.bdc.wfself.entity.TaskRunInfo;
import com.cesgroup.bdc.wfself.entity.TaskRunPartic;
import com.cesgroup.bdc.wfself.node.event.EndEvent;
import com.cesgroup.bdc.wfself.node.event.StartEvent;
import com.cesgroup.bdc.wfself.node.gateway.ExcluGateway;
import com.cesgroup.bdc.wfself.node.gateway.ParaGateway;
import com.cesgroup.bdc.wfself.node.task.UserTask;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeSet;
import java.util.UUID;


/**
 * 流程处理服务
 * 
 * @author khaoth
 *
 */
public abstract class ProcHandleService {

    /**
     * 操作实体
     */
    private WorkflowDao workflowDao;

    /**
     * 流程定义集合
     */
    private Map<String, CommonProcDef> procDefMap;

    /**
     * 设置流程定义集合
     *
     * @param procDefMap
     */
    public void setProcDef(Map<String, CommonProcDef> procDefMap) {
        this.procDefMap = procDefMap;
    }

    /**
     * 获取流程定义集合
     *
     * @return
     */
    public Map<String, CommonProcDef> getProcDef() {
        return this.procDefMap;
    }

    /**
     * 获取指定流程定义
     *
     * @return
     */
    protected CommonProcDef getProcDef(String procDefId) {
        return procDefMap.get(procDefId);
    }

    /**
     * 设置操作实体dao
     *
     * @param workflowDao
     */
    public void setWorkflowDao(WorkflowDao workflowDao) {
        this.workflowDao = workflowDao;
    }

    /**
     * 获取操作实体dao
     *
     * @return
     */
    public WorkflowDao getWorkflowDao() {
        return this.workflowDao;
    }

    /**
     * 获取当前时间，格式yyyy-MM-dd HH:mm:ss
     *
     * @return
     */
    public String getTime() {
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
        return sdf.format(new Date());
    }

    /**
     * 流程驱动
     *
     * @param execId
     *            执行id
     * @param node
     *            处理节点
     * @param userId
     *            当前用户
     * @param paramMap
     *            流程参数
     */
    protected void doHandle(String execId, INode node, String userId, Map<String, String> paramMap) throws Exception {
        if (node instanceof UserTask) {
            this.userTaskHandle(execId, (UserTask)node, userId, paramMap);
        } else if (node instanceof ExcluGateway) {
            this.excluGatewayHandle(execId, (ExcluGateway)node, userId, paramMap);
        } else if (node instanceof ParaGateway) {
            this.paraGatewayHandle(execId, (ParaGateway)node, userId, paramMap);
        } else if (node instanceof EndEvent) {
            this.EndEventHandle(execId, (EndEvent)node, userId);
        } else {
            throw new Exception("没有对应值执行的节点方法， node：" + node.toString());
        }
    }

    /**
     * 开始节点处理
     *
     * @param procDefId
     *            流程定义id
     * @param startEvent
     *            开始节点对象
     * @param userId
     *            用户id
     * @param paramMap
     *            流程参数
     * @return
     */
    protected String startEventHandle(String procDefId, StartEvent startEvent, String userId,
        Map<String, String> paramMap) throws Exception {

        /* 开始节点数据存入历史 */
        String procInstId = UUID.randomUUID().toString().replace("-", "");
        String execId = UUID.randomUUID().toString().replace("-", "");
        String startTime = getTime();
        addProcInstHisInfo(procDefId, startEvent, procInstId, execId, startTime, userId, "");
        /* 获取开始节点的下一个节点，并调用通用处理方法 */
        // 获取流程定义对象
        CommonProcDef commonProcDef = procDefMap.get(procDefId);
        // 获取开始节点的下一个节点
        CommonNode nextNode = (CommonNode)commonProcDef.getProcDefMap().get(startEvent.getCode());
        // 通用处理
        doHandle(execId, nextNode, userId, paramMap);
        return procInstId;
    }

    /**
     * 人工任务节点
     *
     * 如果是串行任务未全完成，需要传入串行任务节点，如果是并行任务未全完成传入 null
     *
     * @param execId
     *            当前节点执行id
     * @param userTask
     *            任务节点
     * @param userId
     *            用户id
     * @param paramMap
     *            节点参数
     */
    private void userTaskHandle(String execId, UserTask userTask, String userId, Map<String, String> paramMap) {

        // 完成当前任务，并返回历史流程实例对象
        ProcInstHisInfo procInstHisInfo = completeCurrTask(execId, userId);

        // 如果没有转任务节点，说明存在并行任务还没完成或其他只要完成任务的情况
        if (userTask != null) {
            // 上一个已完成的流程节点id
            String preNodeId = procInstHisInfo.getId();
            // 上一个已完成的流程节点code
            String preNodeCode = procInstHisInfo.getNodeCode();
            // 获取流程定义id
            String procDefId = procInstHisInfo.getProcDefId();
            // 获取流程实例id
            String procInstId = procInstHisInfo.getProcInstId();
            // 流转的节点code
            String nodeCode = userTask.getCode();

            if (paramMap != null) {
                // 根据上个流程节点code，获取_in_para_gateway流程参数，如果存在，设置给下个任务节点的流程参数
                ProcInstRunVar procInstRunVar =
                    workflowDao.queryProcInstRunVar(procInstId, preNodeCode, ProcConst._IN_PARA_GATEWAY);
                if (procInstRunVar != null) {
                    paramMap.put(ProcConst._IN_PARA_GATEWAY, procInstRunVar.getData());
                }
            }

            /* 创建任务处理 */

            // 如果上一个完成的任务和本次创建任务时相同节点，说明正在执行串行任务
            if (!preNodeCode.equals(nodeCode)) {
                // 处理参与人，返回参与人对应任务id，如果没有paramMap返回null
                List<String> particIdList = taskRunParticHandle(procInstId, userTask, paramMap);

                // 无论串行并行，第一次在此创建任务
                if (particIdList != null && !particIdList.isEmpty()) {
                    String startTime = getTime();
                    for (String particId : particIdList) {
                        String id = addTaskRunInfo(procDefId, userTask, procInstId, startTime, particId);
                        addProcInstHisInfo(procDefId, userTask, procInstId, id, startTime, particId, preNodeId);
                    }
                }

                // 增加流程变量
                procInstRunVarHandle(procInstId, nodeCode, paramMap);

                // 创建剩余串行任务
            } else {

                List<TaskRunPartic> particList = workflowDao.queryTaskRunParticList(procInstId, userTask.getCode());
                if (particList != null && !particList.isEmpty()) {
                    TaskRunPartic taskRunPartic = particList.get(0);
                    String startTime = getTime();
                    String particId = taskRunPartic.getParticId();
                    String id = addTaskRunInfo(procDefId, userTask, procInstId, startTime, particId);
                    // 这里的上一个节点时上一个完成的串行任务
                    addProcInstHisInfo(procDefId, userTask, procInstId, id, startTime, particId, preNodeId);
                    // 删除参与人
                    workflowDao.remove(taskRunPartic);
                }
            }
        }
    }

    /**
     * 排他网关节点处理
     *
     * @param execId
     *            当前节点执行id
     * @param excluGateway
     *            排他网关节点
     * @param userId
     *            用户id
     * @param paramMap
     *            节点参数
     */
    private void excluGatewayHandle(String execId, ExcluGateway excluGateway, String userId,
        Map<String, String> paramMap) throws Exception {

        // 根据执行id获取当前历史任务信息
        ProcInstHisInfo procInstHisInfo = workflowDao.queryProcInstHisInfo(execId);
        // 获取流程定义对象
        CommonProcDef commonProcDef = procDefMap.get(procInstHisInfo.getProcDefId());
        Map<String, Object> procDefMap = commonProcDef.getProcDefMap();
        // 获取网关对应的流程节点集合
        List<INode> nodeList = (List<INode>)procDefMap.get(excluGateway.getCode());
        // 定义的网关流程节点流向key值
        Map<String, String> gatewayFlowKeyMap = commonProcDef.getGatewayFlowKeyMap();
        String key = gatewayFlowKeyMap.get(excluGateway.getCode());
        // 根据排他网关条件获取任务节点
        INode nextNode = excluGateway.condition(paramMap.get(key), nodeList);
        if (nextNode == null) {
            throw new Exception("排他网关无流转节点！");
        }
        // 驱动流程 执行下一步具体流程节点
        doHandle(execId, nextNode, userId, paramMap);
    }

    /**
     * 并行网关节点处理
     *
     * @param execId
     *            当前节点执行id
     * @param paraGateway
     *            并行网关节点
     * @param userId
     *            用户id
     * @param paramMap
     *            节点参数
     */
    private void paraGatewayHandle(String execId, ParaGateway paraGateway, String userId, Map<String, String> paramMap)
        throws Exception {

        ProcInstHisInfo procInstHisInfo = workflowDao.queryProcInstHisInfo(execId);

        // 流程实例id
        String procInstId = procInstHisInfo.getProcInstId();

        // 并行网关的节点code
        String paraGateWayCode = paraGateway.getCode();

        // 获取多个并行节点
        CommonProcDef commonProcDef = procDefMap.get(procInstHisInfo.getProcDefId());
        Map<String, Object> procDefMap = commonProcDef.getProcDefMap();

        // 为分支的并行网关时
        if (!paraGateway.isMerge()) {
            List<INode> nodeList = (List<INode>)procDefMap.get(paraGateWayCode);
            for (INode node : nodeList) {

                // 复制一份参数
                Map<String, String> paramMapTmp = new HashMap<>(paramMap.size());
                paramMapTmp.putAll(paramMap);

                // 判断任务，如果是从并行网关中进入的，增加标记
                if (node instanceof UserTask) {
                    paramMapTmp.put(ProcConst._IN_PARA_GATEWAY, paraGateWayCode);
                }

                this.doHandle(execId, node, userId, paramMapTmp);
            }

            // 并行网关（聚合）
        } else {

            // 有多个任务进入并行网关（聚合），锁住
            synchronized (ProcHandleService.class) {
                // 从并行网关(聚合)中 获取并行网关（分支）code
                String branshCode = paraGateway.getBranshCode();

                // 获取指定实例的流程参数
                List<ProcInstRunVar> procInstRunVarList =
                    workflowDao.queryProcInstRunVarList(procInstId, ProcConst._IN_PARA_GATEWAY, branshCode);

                if (procInstRunVarList != null && !procInstRunVarList.isEmpty()) {

                    TreeSet<String> nodeCodeSet = new TreeSet<>();
                    for (ProcInstRunVar procInstRunVar : procInstRunVarList) {
                        nodeCodeSet.add(procInstRunVar.getNodeCode());
                    }

                    boolean result = true;
                    // 获取流程节点后（此处只可能是人工任务节点）判断是否并行任务都已完成，如果完成可以并行网关（聚合）的下个节点
                    for (String nodeCode : nodeCodeSet) {
                        result &= workflowDao.isUniqueOrCompleteUserTask(procInstId, execId, nodeCode);
                    }

                    // 并行网关（聚合）的下个节点
                    INode nextNode = null;
                    // 为true说明，并行网关中的所有任务只有一个未完成，可以获取下个任务节点，否则传一个null的nextNode
                    if (result) {
                        nextNode = (INode)procDefMap.get(paraGateway.getCode());
                    }
                    doHandle(execId, nextNode, userId, paramMap);
                } else {
                    throw new Exception("流程参数异常！");
                }
            }
        }
    }

    /**
     * 结束节点处理
     *
     * 只要是到了流程结束的接到哪，除了保留历史记录，其他一律删除。
     *
     * @param execId
     *            执行id
     * @param userId
     *            用户id
     * @param endEvent
     *            结束节点
     */
    private void EndEventHandle(String execId, EndEvent endEvent, String userId) {

        // 完成当前任务，并返回历史流程实例对象
        ProcInstHisInfo procInstHisInfo = completeCurrTask(execId, userId);
        String procDefId = procInstHisInfo.getProcDefId();
        String procInstId = procInstHisInfo.getProcInstId();
        // 结束时间执行id
        String endEventExecId = UUID.randomUUID().toString().replace("-", "");
        // 上一个节点id，不会是网关的id
        String preNodeId = procInstHisInfo.getId();
        String startTime = getTime();
        // 创建历史流程任务
        addProcInstHisInfo(procDefId, endEvent, procInstId, endEventExecId, startTime, userId, preNodeId);
        // 完成任务
        completeCurrTask(endEventExecId, userId);

        /* ----------------------删除无用数据 --------------------*/

        // 删除流程运行变量
        workflowDao.removeAll(workflowDao.queryProcInstRunVarList(procInstId));
        /* 删除运行时任务 */
        List<TaskRunInfo> taskRunInfoList = workflowDao.queryTaskRunInfoList(procInstId);
        if(taskRunInfoList != null && !taskRunInfoList.isEmpty()) {
            workflowDao.removeAll(taskRunInfoList);
            /* 删除未完成的历史流程记录 */
            List<String> execIdList = new ArrayList<>();
            for(TaskRunInfo taskRunInfo : taskRunInfoList) {
                execIdList.add(taskRunInfo.getId());
            }
            workflowDao.removeAll(workflowDao.queryProcInstHisInfoList(execIdList));
        }
        // 删除参与人数据
        workflowDao.removeAll(workflowDao.queryTaskRunParticList(procInstId));

    }

    /**
     * 参与人处理
     *
     * 规则，如果是候选人，无论串行、并行返回一个参与者id。 如果是委托人，串行返回一个参与者id，并行返回所有参与人id
     *
     * 返回的参与人id不存入参与者表中
     *
     * @param procInstId
     *            流程实例id
     * @param userTask
     *            人工任务节点
     * @param paramMap
     *            流程参数
     */
    private List<String> taskRunParticHandle(String procInstId, UserTask userTask, Map<String, String> paramMap) {

        if (paramMap != null && !paramMap.isEmpty()) {
            // 参与人数据
            String[] particArray = null;
            // 获取委托人id list
            String assigneeList = paramMap.get(ProcConst.PARTIC_ASSIGNEE_LIST);
            // 获取候选人id list
            String candidateList = paramMap.get(ProcConst.PARTIC_CANDIDATE_LIST);
            // 返回的应参与人id， 并且如果是候选人类型，对应人为空字符串
            List<String> resultList = new ArrayList<>();
            // 参与人类型
            String particType = null;

            // 判断是否存在委托人
            if (assigneeList != null && !"".equals(assigneeList)) {
                particArray = assigneeList.split(",");
                particType = ProcConst.PARTIC_TYPE_ASSIGNEE;

                // 判断是否存在候选人
            } else if (candidateList != null && !"".equals(candidateList)) {
                particArray = candidateList.split(",");
                particType = ProcConst.PARTIC_TYPE_CANDIDATE;
            }

            // 参与人入库
            if (particArray != null && particArray.length > 0) {

                // 控制返回的参与人id
                boolean flag = true;

                // 循环遍历委托人的值
                for (int i = 0; i < particArray.length; i++) {

                    // 参与人id
                    String particId = particArray[i];

                    // 如果是委托人类型
                    if (ProcConst.PARTIC_TYPE_ASSIGNEE.equals(particType)) {

                        // 串行任务
                        if (flag && ITask.JOINT_SIGN_SERIAL.equals(userTask.getJointSign())) {
                            resultList.add(particId);
                            flag = false;
                            continue;
                        }

                        // 并行任务，不存入参与者表
                        if (ITask.JOINT_SIGN_SERIAL.equals(userTask.getJointSign())) {
                            resultList.add(particId);
                            continue;
                        }

                        // 如果是候选人类型，意味着抢占式任务，设置一个控制，意味着只创建一个运行时任务和历史数据。
                    } else if (flag && ProcConst.PARTIC_TYPE_CANDIDATE.equals(particType)) {
                        resultList.add("");
                        flag = false;
                    }

                    // 参与者对象赋值
                    TaskRunPartic taskRunPartic = new TaskRunPartic();
                    taskRunPartic.setProcInstId(procInstId);
                    taskRunPartic.setParticId(particId);
                    taskRunPartic.setOrderNum(i + "");
                    taskRunPartic.setType(particType);
                    taskRunPartic.setTaskCode(userTask.getCode());
                    workflowDao.save(taskRunPartic);
                }
            }
            return resultList;
        }
        return null;
    }

    /**
     * 流程变量处理
     *
     * 与流程节点绑定，如果此节点有新的参数，则删除原有的，插入新的
     *
     * @param procInstId
     *            流程实例id
     * @param nodeCode
     *            流程节点
     * @param paramMap
     *            流程参数
     */
    private void procInstRunVarHandle(String procInstId, String nodeCode, Map<String, String> paramMap) {

        if (paramMap != null && !paramMap.isEmpty()) {

            List<ProcInstRunVar> procInstRunVarList = workflowDao.queryProcInstRunVarList(procInstId, nodeCode);
            // 如果不等于空，删除原有数据
            if (procInstRunVarList != null && !procInstRunVarList.isEmpty()) {
                workflowDao.removeAll(procInstRunVarList);
            }

            Iterator<String> iterator = paramMap.keySet().iterator();
            while (iterator.hasNext()) {
                String key = iterator.next();
                /* 开始节点数据存入历史 */
                ProcInstRunVar procInstRunVar = new ProcInstRunVar();
                procInstRunVar.setProcInstId(procInstId);
                procInstRunVar.setNodeCode(nodeCode);
                procInstRunVar.setName(key);
                procInstRunVar.setData(paramMap.get(key));
                workflowDao.save(procInstRunVar);
            }
        }
    }

    /**
     * 新增运行流程实例数据，并返回主键id
     *
     * @param procDefId
     *            流程定义id
     * @param userTask
     *            流程节点
     * @param procInstId
     *            流程实例id
     * @param startTime
     *            开始时间，如果节点相同，开始时间一致，说明时并行任务
     * @param userId
     *            用户id，如果参与人类型时候选人，userId应该为空串
     */
    private String addTaskRunInfo(String procDefId, UserTask userTask, String procInstId, String startTime,
        String userId) {
        /* 开始节点数据存入运行流程实例表 */
        TaskRunInfo taskRunInfo = new TaskRunInfo();
        taskRunInfo.setProcDefId(procDefId);
        taskRunInfo.setProcInstId(procInstId);
        taskRunInfo.setTaskCode(userTask.getCode());
        taskRunInfo.setTaskName(userTask.getName());
        taskRunInfo.setStartTime(startTime);
        taskRunInfo.setAssgineeId(userId);
        workflowDao.save(taskRunInfo);
        return taskRunInfo.getId();
    }

    /**
     * 新增历史流程实例数据
     *
     * @param procDefId
     *            流程定义id
     * @param node
     *            流程节点
     * @param procInstId
     *            流程实例id
     * @param execId
     *            执行id
     * @param startTime
     *            开始时间
     * @param userId
     *            用户id
     * @param preNodeId
     *            上一个流程节点id
     */
    private void addProcInstHisInfo(String procDefId, CommonNode node, String procInstId, String execId,
        String startTime, String userId, String preNodeId) {
        /* 开始节点数据存入历史 */
        ProcInstHisInfo procInstHisInfo = new ProcInstHisInfo();
        String userName = workflowDao.getNameById(userId);
        String orgId = workflowDao.getOrgIdByUserId(userId);
        procInstHisInfo.setProcDefId(procDefId);
        procInstHisInfo.setProcInstId(procInstId);
        procInstHisInfo.setExecId(execId);
        procInstHisInfo.setNodeCode(node.getCode());
        procInstHisInfo.setNodeName(node.getName());
        procInstHisInfo.setPreNodeId(preNodeId);
        procInstHisInfo.setStartTime(startTime);
        procInstHisInfo.setAssgineeId(userId);
        procInstHisInfo.setAssgineeName(userName);
        procInstHisInfo.setAssgineeOrgId(orgId);
        workflowDao.save(procInstHisInfo);
    }

    /**
     * 完成当前任务 并返回历史流程实例对象
     *
     * @param execId
     *            执行id
     * @param userId
     *            处理人id
     */
    protected ProcInstHisInfo completeCurrTask(String execId, String userId) {

        ProcInstHisInfo procInstHisInfo = workflowDao.queryProcInstHisInfo(execId);

        // 结束时间
        String endTime = procInstHisInfo.getEndTime();
        // 如果没有结束时间
        if (endTime == null || "".equals(endTime)) {
            String userName = workflowDao.getNameById(userId);
            String orgId = workflowDao.getOrgIdByUserId(userId);
            procInstHisInfo.setStatus(ProcConst.STATUS_COMPLETED);
            procInstHisInfo.setEndTime(getTime());
            procInstHisInfo.setHandlerId(userId);
            procInstHisInfo.setHandlerName(userName);
            procInstHisInfo.setHandlerOrgId(orgId);
            workflowDao.update(procInstHisInfo);
            workflowDao.removeRunTaskInfo(execId);
        }
        return procInstHisInfo;
    }

}
